/*		 ICP-PCI Kernel module free public release
 *
 *		       Copyright (C) 1996-99 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 *
 * $Id: icpciOS.h,v 5.1.2.2 1999/07/13 07:57:38 matthias Exp $
 *
 */

#ifndef _ICPCIOS_H
#define _ICPCIOS_H

/*
 * Includes:
 */

/* #undef linux */
#undef OS_DEFINED

#ifdef linux
#  define OS_DEFINED
#  include <linux/types.h>	/* all kernel data types */
#  include <linux/string.h>	/* memcpy, memset, ... */
#  include <linux/kernel.h>	/* printk */
#  include <linux/errno.h>	/* ENXIO, EINVAL, ... */
#  include <linux/version.h>	/* LINUX_VERSION_CODE */
#  include <linux/malloc.h>	/* kmalloc */
#  include <linux/time.h>	/* do_gettimeofday */
#  include <linux/timer.h>	/* timer_list */
#  include <linux/wait.h>	/* wait_queue */
#  include <linux/sched.h>	/* task_struct */
#  include <linux/tqueue.h>	/* tq_struct */
#  include <linux/interrupt.h>	/* mark_bh */
#  include <linux/signal.h>	/* SIGIO */
#  include <linux/mm.h>		/* vremap, PAGE_SIZE */
#  include <linux/kdev_t.h>	/* MAJOR / MINOR */
#  include <linux/major.h>	/* MAX_CHRDEV */
#  include <linux/delay.h>	/* udelay */
#  include <linux/pci.h>	/* pci specific stuff */
#  include <asm/irq.h>		/* enable_irq */
#  include <asm/io.h>		/* readl, outb... */
#  if LINUX_VERSION_CODE < 0x020100
#    include <asm/segment.h>	/* memcpy_tofs */
#    include <linux/bios32.h>	/* pci bios access routines */
#  elif LINUX_VERSION_CODE > 0x020200
#    include <linux/poll.h>	/* poll_table */
#    include <linux/vmalloc.h>	/* vmalloc */ 
#    include <asm/uaccess.h>	/* copy_to_user */
#  endif
#endif

/*
 * Defines:
 */

#ifdef linux
#  define ICP_PCI_BASE_ADDR_SPACE		PCI_BASE_ADDRESS_SPACE
#  define ICP_PCI_BASE_ADDR_SPACE_MEMORY	PCI_BASE_ADDRESS_SPACE_MEMORY
#  define ICP_PCI_BASE_ADDR_SPACE_IO		PCI_BASE_ADDRESS_SPACE_IO
#  define ICP_PCI_BASE_ADDR_MEM_MASK		PCI_BASE_ADDRESS_MEM_MASK
#  define ICP_PCI_BASE_ADDR_IO_MASK		PCI_BASE_ADDRESS_IO_MASK
#  if LINUX_VERSION_CODE < 0x020100
#    define ICP_SEL_IN	SEL_IN
#    define ICP_SEL_EX	SEL_EX
#  elif LINUX_VERSION_CODE > 0x020200
#    define ICP_SEL_IN	1
#    define ICP_SEL_EX	4
#  endif
#  define ICP_MAJOR_0	60	/* major device number of fg board -
				   first entry for local (experimental) use */
#  define ICP_MAJOR_1	61	/* since we have more than 6 acquisition
				   modules, we need another major number */
#endif

#define ICP_SIGIO	SIGIO

#ifndef OS_DEFINED /* if nothing is defined, fallback */
#  define	EINTR	 1	/* Interrupted system call */
#  define	EIO	 2	/* I/O error */
#  define	ENXIO	 3	/* No such device or address */
#  define	EAGAIN	 4	/* Try again */
#  define	ENOMEM	 5	/* Out of memory */
#  define	EFAULT	 6	/* Bad address */
#  define	EBUSY	 7	/* Device or resource busy */
#  define	EINVAL	 8	/* Invalid argument */
#  define	EFBIG	 9	/* File too large */
#  define	ENODATA	10	/* No data available */
#  define PAGE_SIZE	(1 << 12)	/* 4kB */
#  define HZ		100		/* 10ms */
#  define O_EXCL		0001
#  define O_NONBLOCK	0002
#  define ICP_SEL_IN	0001
#  define ICP_SEL_EX	0004
#  define ICP_PCI_BASE_ADDR_SPACE		0x01
#  define ICP_PCI_BASE_ADDR_SPACE_MEMORY	0x00
#  define ICP_PCI_BASE_ADDR_SPACE_IO	0x01
#  define ICP_PCI_BASE_ADDR_MEM_MASK	(~0x0fUL)
#  define ICP_PCI_BASE_ADDR_IO_MASK	(~0x03UL)
#  define ICP_MAJOR_0	0
#  define ICP_MAJOR_1	1
#endif

/*
 * Typedefs:
 */

#ifdef linux
#  if LINUX_VERSION_CODE < 0x020100
typedef u_short icp_flags_t;
typedef int icp_size_t;
typedef int icp_ssize_t;
#  elif LINUX_VERSION_CODE > 0x020200
typedef u_int icp_flags_t;
typedef size_t icp_size_t;
typedef ssize_t icp_ssize_t;
#  endif
#  if LINUX_VERSION_CODE < 0x020020
typedef int icp_f_owner_t;
#  else
typedef struct fown_struct icp_f_owner_t;
#  endif
#endif

/* if nothing is defined, fallback */
#ifndef OS_DEFINED
#  define u_char	unsigned char
#  define u_short	unsigned short
#  define u_long	unsigned long
#  define u_int		unsigned int
typedef long	off_t;
typedef u_long	time_t;
typedef u_long	size_t;
typedef u_long	ssize_t;
typedef long	ptrdiff_t;
typedef size_t icp_size_t;
typedef ssize_t icp_ssize_t;
typedef u_long icp_flags_t;
typedef u_long icp_f_owner_t;
#endif

#ifdef linux
struct icp_memset_t
{
  int board;
  void *addr;
  char value;
  size_t size;
};
#endif

struct icp_osp_t
{
#ifdef linux
  u_long flags;	/* for the inline splhigh / spllow routines */
  struct wait_queue *cam_w_queue;	/* queue to wait for camera events */
  struct wait_queue *iop_w_queue;	/* queue to wait for io-port events */
  struct timer_list cam_timer;	/* timeout for camera wait queues */
  struct timer_list iop_timer;	/* timeout for io-port wait queues */
  struct tq_struct clr_t_queue;	/* struct to add the mem clear bottom half */
#  if LINUX_VERSION_CODE < 0x020100
  int signum;	/* signal to send on i/o completion */
#    if LINUX_VERSION_CODE < 0x020020	/* 2.0.32 */
  int pid;	/* send this signal to pid # */	
#    else
  struct fown_struct fown;
#    endif
#  else
  struct fown_struct fown;
#  endif
  struct task_struct *task;
  struct icp_memset_t memset;	/* for delayed memset (task_queues) */
#endif
};

/*
 * Prototypes:
 */

int
icp_printf (const char*, ...);
int
icp_irq_printf (const char*, ...);
void*
icp_malloc (size_t, int);
void
icp_free (void*, int);
int
icp_add_timer (int, time_t, int);
int
icp_del_timer (int, int);
int
icp_sleep (int, time_t, int);
int
icp_wakeup (int, int);
int
icp_marktq (int);
void
icp_irqoff (int);
void
icp_irqon (int);
void
icp_splhigh (int);
void
icp_spllow (int);
int
icp_fill_info (int, void*);
int
icp_assign_task (int);
int
icp_deassign_task (int);
int
icp_task_assigned (int);
int
icp_assign_signal (int, icp_flags_t, icp_f_owner_t*);
int
icp_deassign_signal (int);
int
icp_check_blocked (int);
int
icp_kill_proc (int);

int
icp_os_init (int);
int
icp_os_pci_init (int*);
int
icp_os_map_init (int, u_long*, ptrdiff_t*);
int
icp_os_cap_init (int);

u_short
icp_readw (u_short*);
void
icp_writew (u_short, u_short*);
u_long
icp_readl (u_long*);
void
icp_writel (u_long, u_long*);
u_short
icp_inw (u_short);
void
icp_outw (u_short, u_short);
u_long
icp_inl (u_short);
void
icp_outl (u_long, u_short);
void*
icp_memset_io (int, void*, char, size_t);
void*
icp_memcpy_fromio (void*, void*, size_t);
u_long
icp_virt_to_bus (void*);
void
icp_copyout (void*, void*, size_t);
void
icp_copyin (void*, const void*, size_t);
void
icp_udelay (u_long);

#ifdef linux
#  define icp_printf(fmt,args...) printk (KERN_DEBUG fmt, ## args)
#  define icp_irq_printf(fmt,args...) printk (fmt, ## args)
extern inline u_short
icp_readw (u_short *addr) 
{
  return readw (addr);
}
extern inline void
icp_writew (u_short value, u_short *addr)
{
  writew (value, addr);
}
extern inline u_long
icp_readl (u_long *addr) 
{
  return readl (addr);
}
extern inline void
icp_writel (u_long value, u_long *addr)
{
  writel (value, addr);
}
extern inline u_short
icp_inw (u_short port) 
{
  return inw (port);
}
extern inline void
icp_outw (u_short value, u_short port)
{
  outw (value, port);
}
extern inline u_long
icp_inl (u_short port) 
{
  return inl (port);
}
extern inline void
icp_outl (u_long value, u_short port)
{
  outl (value, port);
}
extern inline void*
icp_memcpy_fromio (void *dst, void *src, size_t size)
{
  return memcpy_fromio (dst, src, size);
}
extern inline u_long
icp_virt_to_bus (void *addr)
{
  return virt_to_bus (addr);
}
extern inline void
icp_udelay (u_long time)
{
  udelay (time);
}

#if LINUX_VERSION_CODE < 0x020100
extern inline void
icp_copyout (void *dst, void *src, size_t size)
{
  memcpy_tofs (dst, src, size);
}
extern inline void
icp_copyin (void *dst, const void *src, size_t size)
{
  memcpy_fromfs (dst, src, size);
}
#elif LINUX_VERSION_CODE > 0x020200
extern inline void
icp_copyout (void *dst, void *src, size_t size)
{
  copy_to_user (dst, src, size);
}
extern inline void
icp_copyin (void *dst, const void *src, size_t size)
{
  copy_from_user (dst, src, size);
}
#endif
#endif /* linux */

#ifndef OS_DEFINED
/* void* memcpy (void*, const void*, size_t); */
void* memset (void*, int, size_t);
#endif

#endif /* _ICPCIOS_H */

