/*		 ICP-PCI Kernel module free public release
 *
 *		       Copyright (C) 1996-99 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: icpciReg.c,v 5.1.2.5 1999/08/27 15:40:28 matthias Exp $
 *
 */

/*
 * Defines:
 */

/* #define DEBUG_ICPCI_REG */

/*
 * Includes:
 */

#include "icpciOS.h"	/* os specific stuff */
#include "icpciExt.h"	/* offical driver interface */
#include "icpciInt.h"	/* driver internals */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "ampcvReg.h"	/* register definitions PCVision
			   (for interrupt test/clear functions) */

/*
 * Globals:
 */

extern struct
icp_global_t    icp;	/* main driver structure */

#ifdef DEBUG_ICPCI_REG
const char *icr_reg[] = { "mbox1 ",
			  "mbox2 ",
			  "mbox3 ",
			  "mbox4 ",
			  "error ",
			  "error ",
			  "error ",
			  "error ",
			  "error ",
			  "bmdst ",
			  "bmxc  ",
			  "error ",
			  "error ",
			  "error ",
			  "intctl",
			  "bmctl "
			};

const char  *cr_reg[] = { "con   ",
			  "acqcon",
			  "pitch ",
			  "aoix  ",
			  "bmsfl ",
			  "bmsfh ",
			  "bmssl ",
			  "bmssh ",
			  "aqstrt",
			  "aqadr ",
			  "intadr",
			  "intcon",
			  "wzoom ",
			  "mdata ", 
			  "error ",
			  "error ", 
			  "prtcon",
			  "outprt",
			  "inport" 
			};
#endif /* DEBUG_ICPCI_REG */

/*
 * Prototypes:
 */

/*
 * General Read Handler for IC-PCI Interface Control Registers
 */

/* 32 bit register access with mask and range checking */

static u_long
icr_read_mm (board, offset, mask, min, max)
     
  u_char board;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_long mask;		/* read mask for register */
  u_long min;		/* minimal value for range checking */
  u_long max;		/* maximal value for range checking */

{
  u_long tmp_val = 0;

  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait32);
    }
    tmp_val = icp_readl (icp.map[board]->v_icr_base + offset /
			 sizeof(u_long));
  } else {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait32);
    }
    tmp_val = icp_inl (icp.map[board]->io_icr_base + offset);
  }
  
  tmp_val &= mask;

#if(0)
  if (tmp_val & ~mask) icp.errno[board] = -EICP_MASK;

  if (tmp_val < min || tmp_val > max) icp.errno[board] = -EICP_RANGE;
#endif

#ifdef DEBUG_ICPCI_REG
  if (!icp.map[board]->regs_io)
    icp_printf("icpci[%d]: icr-readl  "
	       "%s, %#010lX &(%#010lX) >%#010lX <%#010lX\n",
	       board, icr_reg[offset / sizeof(u_long)],
	       tmp_val, mask, min, max);
  else
    icp_printf("icpci[%d]: icr-inl    "
	       "%s, %#010lX &(%#010lX) >%#010lX <%#010lX\n",
	       board, icr_reg[offset / sizeof(u_long)],
	       tmp_val, mask, min, max);
#endif

  return (tmp_val);
}

/* 32 bit register 1 (or more) bit access */

static u_char
icr_read_ms (board, offset, mask, shift)

  u_char board;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_char shift;		/* number of bits to shift right */
  u_long mask;		/* read mask for register */
  
{
  u_long tmp_val = 0;

  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait32);
    }
    tmp_val = icp_readl (icp.map[board]->v_icr_base + offset /
			 sizeof(u_long));
  } else {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait32);
    }
    tmp_val = icp_inl (icp.map[board]->io_icr_base + offset);
  }

  tmp_val &= mask;
  tmp_val >>= shift;

#ifdef DEBUG_ICPCI_REG
  if (!icp.map[board]->regs_io)
    icp_printf("icpci[%d]: icr-readl  %s, %#1lX &(%#010lX) >>%2d\n",
	       board, icr_reg[offset / sizeof(u_long)], tmp_val, mask, shift);
  else
    icp_printf("icpci[%d]: icr-inl    %s, %#1lX &(%#010lX) >>%2d\n",
	       board, icr_reg[offset / sizeof(u_long)], tmp_val, mask, shift);
#endif

  return ((u_char)tmp_val);
}
  
/*
 * General Write Handler for Interface Control Registers
 */

/* 32 bit register access with mask and range checking */

static int
icr_write_mm (board, offset, mask, min, max, value)

  u_char board;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_long mask;		/* read mask for register */
  u_long min;		/* minimal value for range checking */
  u_long max;		/* maximal value for range checking */
  u_long value;		/* value to set to */

{
  u_long tmp_val = 0;

  if (value & ~mask) {
    icp.errno[board] = -EICP_MASK;
    return (ERROR);
  }

  if (value < min || value > max) {
    icp.errno[board] = -EICP_RANGE;  
    return (ERROR);
  }

  tmp_val = value & mask;

#ifdef DEBUG_ICPCI_REG
  if (!icp.map[board]->regs_io)
    icp_printf("icpci[%d]: icr-writel "
	       "%s, %#010lX &(%#010lX) >%#010lX <%#010lX\n",
	       board, icr_reg[offset / sizeof(u_long)],
	       tmp_val, mask, min, max);
  else
    icp_printf("icpci[%d]: icr-outl   "
	       "%s, %#010lX &(%#010lX) >%#010lX <%#010lX\n",
	       board, icr_reg[offset / sizeof(u_long)],
	       tmp_val, mask, min, max);
#endif

  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait32);
    }
    icp_writel (tmp_val, icp.map[board]->v_icr_base + offset /
		sizeof(u_long));
  } else {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait32);
    }
    icp_outl (tmp_val, icp.map[board]->io_icr_base + offset);
  }

  return (OK);
}

/* 32 bit register 1 (or more) bit access */

static int
icr_write_ms (board, offset, mask, shift, value)

  u_char board;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_char shift;		/* number of bits to shift right */
  u_char value;		/* value to set to */  
  u_long mask;		/* read mask for register */
  
{
  u_long read_val = 0;
  u_long tmp_val = 0;

  tmp_val = (u_long)value << shift;

  if (tmp_val & ~mask) {
    icp.errno[board] = -EICP_MASK;
    return (ERROR);
  }

  /* this have to be atomar */
  icp_splhigh (board);

  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait32);
    }
    read_val = icp_readl (icp.map[board]->v_icr_base + offset /
			  sizeof(u_long));
  } else {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait32);
    }
    read_val = icp_inl (icp.map[board]->io_icr_base + offset);
  }
  
  read_val &= ~mask;
  read_val |= tmp_val;

#ifdef DEBUG_ICPCI_REG
  if (!icp.map[board]->regs_io)
    icp_printf("icpci[%d]: icr-writel %s, %#010lX &(%#010lX) <<%2d\n",
	       board, icr_reg[offset / sizeof(u_long)], tmp_val, mask, shift);
  else
    icp_printf("icpci[%d]: icr-outl   %s, %#010lX &(%#010lX) <<%2d\n",
	       board, icr_reg[offset / sizeof(u_long)], tmp_val, mask, shift);
#endif

  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait32);
    }
    icp_writel (read_val, icp.map[board]->v_icr_base + offset /
		sizeof(u_long));
  } else {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait32);
    }
    icp_outl (read_val, icp.map[board]->io_icr_base + offset);
  }

  icp_spllow (board); /* end atomar */

  return (OK);
}
  
/*
 * Special Read Handler for IC-PCI Interface Control Registers
 */

u_long
icr_get_mbox1 (u_char board) {

  return (icr_read_mm (board,
		       ICR_MBOX1_OFFSET,
		       ICR_MBOX_MASK,
		       ICR_MBOX_MIN,
		       ICR_MBOX_MAX));
}

u_long
icr_get_mbox2 (u_char board) {

  return (icr_read_mm (board,
		       ICR_MBOX2_OFFSET,
		       ICR_MBOX_MASK,
		       ICR_MBOX_MIN,
		       ICR_MBOX_MAX));
}

u_long
icr_get_mbox3 (u_char board) {

  return (icr_read_mm (board,
		       ICR_MBOX3_OFFSET,
		       ICR_MBOX_MASK,
		       ICR_MBOX_MIN,
		       ICR_MBOX_MAX));
}

u_long
icr_get_mbox4 (u_char board) {

  return (icr_read_mm (board,
		       ICR_MBOX4_OFFSET,
		       ICR_MBOX_MASK,
		       ICR_MBOX_MIN,
		       ICR_MBOX_MAX));
}

u_long
icr_get_bmdst (u_char board) {

  return (icr_read_mm (board,
		       ICR_BMDST_OFFSET,
		       ICR_BMDST_MASK,
		       ICR_BMDST_MIN,
		       ICR_BMDST_MAX));
}

u_long
icr_get_bmxc (u_char board) {
  
  if (icp.map[board]->big_mem) {
    return (icr_read_mm (board,
			 ICR_BMXC_OFFSET,
			 (ICR_BMXC_MASK << 1) + 1,
			 ICR_BMXC_MIN,
			 (ICR_BMXC_MAX << 1) + 1));
  } else {
    return (icr_read_mm (board,
			 ICR_BMXC_OFFSET,
			 ICR_BMXC_MASK,
			 ICR_BMXC_MIN,
			 ICR_BMXC_MAX));
  }
}

u_char
icr_get_inten (u_char board) {

    return (icr_read_ms (board,
			 ICR_INTCTL_OFFSET,
			 ICR_INTEN_MASK,
			 ICR_INTEN_SHIFT));
}

u_char
icr_get_binten (u_char board) {

  return (icr_read_ms (board,
		       ICR_INTCTL_OFFSET,
		       ICR_INTEN_MASK,
		       ICR_INTEN_SHIFT));
}

/* old, use xcr_test_irq */
u_char
icr_get_intst (u_char board) {

  return (icr_read_ms (board,
		       ICR_INTCTL_OFFSET,
		       ICR_INTST_MASK,
		       ICR_INTST_SHIFT));
}

/* old, use xcr_test_irq */
u_char
icr_get_bintst (u_char board) {

  return (icr_read_ms (board,
		       ICR_INTCTL_OFFSET,
		       ICR_BINTST_MASK,
		       ICR_BINTST_SHIFT));
}

/* old, use xcr_test_irq */
u_char
icr_get_maint (u_char board) {

  return (icr_read_ms (board,
		       ICR_INTCTL_OFFSET,
		       ICR_MAINT_MASK,
		       ICR_MAINT_SHIFT));
}

/* old, use xcr_test_irq */
u_char
icr_get_taint (u_char board) {

  return (icr_read_ms (board,
		       ICR_INTCTL_OFFSET,
		       ICR_TAINT_MASK,
		       ICR_TAINT_SHIFT));
}

u_char
icr_get_fifofl (u_char board) {

  return (icr_read_ms (board,
		       ICR_BMCTL_OFFSET,
		       ICR_FIFOFL_MASK,
		       ICR_FIFOFL_SHIFT));
}

u_char
icr_get_fifo4p (u_char board) {

  return (icr_read_ms (board,
		       ICR_BMCTL_OFFSET,
		       ICR_FIFO4P_MASK,
		       ICR_FIFO4P_SHIFT));
}

u_char
icr_get_fifoem (u_char board) {

  return (icr_read_ms (board,
		       ICR_BMCTL_OFFSET,
		       ICR_FIFOEM_MASK,
		       ICR_FIFOEM_SHIFT));
}

u_char
icr_get_bmdone (u_char board) {

  return (icr_read_ms (board,
		       ICR_BMCTL_OFFSET,
		       ICR_BMDONE_MASK,
		       ICR_BMDONE_SHIFT));
}

u_char
icr_get_bmreq (u_char board) {

  return (icr_read_ms (board,
		       ICR_BMCTL_OFFSET,
		       ICR_BMREQ_MASK,
		       ICR_BMREQ_SHIFT));
}

/*
 * Special Write Handler for IC-PCI Interface Control Registers
 */

int
icr_set_mbox1 (u_char board, u_long value) {

  return (icr_write_mm (board,
			ICR_MBOX1_OFFSET,
			ICR_MBOX_MASK,
			ICR_MBOX_MIN,
			ICR_MBOX_MAX,
			value));
}

int
icr_set_mbox2 (u_char board, u_long value) {

  return (icr_write_mm (board,
			ICR_MBOX2_OFFSET,
			ICR_MBOX_MASK,
			ICR_MBOX_MIN,
			ICR_MBOX_MAX,
			value));
}

int
icr_set_mbox3 (u_char board, u_long value) {

  return (icr_write_mm (board,
			ICR_MBOX3_OFFSET,
			ICR_MBOX_MASK,
			ICR_MBOX_MIN,
			ICR_MBOX_MAX,
			value));
}

int
icr_set_mbox4 (u_char board, u_long value) {

  return (icr_write_mm (board,
			ICR_MBOX4_OFFSET,
			ICR_MBOX_MASK,
			ICR_MBOX_MIN,
			ICR_MBOX_MAX,
			value));
}

int
icr_set_bmdst (u_char board, u_long value) {

  return (icr_write_mm (board,
			ICR_BMDST_OFFSET,
			ICR_BMDST_MASK,
			ICR_BMDST_MIN,
			ICR_BMDST_MAX,
			value));
}

int
icr_set_bmxc (u_char board, u_long value) {

  if (icp.map[board]->big_mem) {
    return (icr_write_mm (board,
			  ICR_BMXC_OFFSET,
			  (ICR_BMXC_MASK << 1) + 1,
			  ICR_BMXC_MIN,
			  (ICR_BMXC_MAX << 1) + 1,
			  value));
  } else {
    return (icr_write_mm (board,
			  ICR_BMXC_OFFSET,
			  ICR_BMXC_MASK,
			  ICR_BMXC_MIN,
			  ICR_BMXC_MAX,
			  value));
  }
}

/* old, use icp_[en|dis]able_irq */
int
icr_set_inten (u_char board, u_char value) {

  return (icr_write_ms (board,
			ICR_INTCTL_OFFSET,
			ICR_INTEN_MASK,
			ICR_INTEN_SHIFT,
			value));
	    
}

/* old, use icp_[en|dis]able_irq */
int
icr_set_binten (u_char board, u_char value) {

  return (icr_write_ms (board,
			ICR_INTCTL_OFFSET,
			ICR_BINTEN_MASK,
			ICR_BINTEN_SHIFT,
			value));
}

int
icr_set_bmreq (u_char board, u_char value) {

  if (value == ICR_ENABLE) {
    return (icr_write_mm (board,
			  ICR_BMCTL_OFFSET,
			  ICR_NO_MASK,
			  ICR_NO_MIN,
			  ICR_NO_MAX,
			  ICR_STRT_TRANS));
  } else {
    return (icr_write_ms (board,
			  ICR_BMCTL_OFFSET,
			  ICR_BMREQ_STOP_MASK,
			  ICR_NO_SHIFT,
			  ICR_STOP_TRANS));
  }
}

/*
 * Special Clear Handler for IC-PCI Interface Control Registers
 */

int
icr_clear_fifoclr (u_char board) {

    return (icr_write_ms (board,
			  ICR_BMCTL_OFFSET,
			  ICR_FIFOCLR_MASK,
			  ICR_FIFOCLR_SHIFT,
			  ICR_ENABLE));
}

/*
 * Special Initialization Handler for Interface Control Registers
 */

int
icr_init_intctl (u_char board) {

    return (icr_write_mm (board,
			  ICR_INTCTL_OFFSET,
			  ICR_NO_MASK,
			  ICR_NO_MIN,
			  ICR_NO_MAX,
			  ICR_INTCTL_INIT));
}
		
int
icr_init_bmctl (u_char board) {

  return (icr_write_mm (board,
			ICR_BMCTL_OFFSET,
			ICR_NO_MASK,
			ICR_NO_MIN,
			ICR_NO_MAX,
			ICR_BMCTL_INIT));
}

/*
 * General Read Handler for IC-PCI Control Registers
 */

/* 16 bit register access with mask and range checking */

static u_short
cr_read_mm (board, offset, mask, min, max)
     
  u_char board;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_short mask;		/* read mask for register */
  u_short min;		/* minimal value for range checking */
  u_short max;		/* maximal value for range checking */

{
  u_short tmp_val = 0;
  u_short amcc_dummy = 0;
  
  if (icp.cap[board]->amcc0_bug) {
    if (icp_check_amcc (board, icp.map[board]->v_cr_base + offset / 
			sizeof(u_short), ICP_AMCC_READ)) {
      if (!icp.map[board]->regs_io) {
	ICP_DELAY(icp.cap[board]->mem_wait16);
	amcc_dummy = icp_readw (icp.map[board]->v_cr_base + offset /
				sizeof(u_short));
      } else {
	ICP_DELAY(icp.cap[board]->io_wait16);
	amcc_dummy = icp_inw (icp.map[board]->io_cr_base + offset);
      }
    }
  }
  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait16);
    }
    tmp_val = icp_readw (icp.map[board]->v_cr_base + offset /
			 sizeof(u_short));
  } else {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait16);
    }
    tmp_val = icp_inw (icp.map[board]->io_cr_base + offset);
  }
  
  tmp_val &= mask;
  
#if(0)
  if (tmp_val & mask) icp.errno[board] = -EICP_MASK;
  
  if (tmp_val < min || tmp_val > max) icp.errno[board] = -EICP_RANGE;
#endif
  
#ifdef DEBUG_ICPCI_REG
  if (!icp.map[board]->regs_io)
    icp_printf("icpci[%d]:  cr-readw  %s, %#06X &(%#06X) >%#06X <%#06X\n",
	       board, cr_reg[offset / sizeof(u_short)],
	       tmp_val, mask, min, max);
  else
    icp_printf("icpci[%d]:  cr-inw    %s, %#06X &(%#06X) >%#06X <%#06X\n",
	       board, cr_reg[offset / sizeof(u_short)],
	       tmp_val, mask, min, max);
#endif
  
  return (tmp_val);
}

/* 16 bit register 1 (or more) bit access */

static u_char
cr_read_ms (board, offset, mask, shift)
     
     u_char board;	/* number of IC-PCI board attached to the computer */
     u_char offset;	/* offset to register base address pointer */
     u_char shift;	/* number of bits to shift right */
     u_short mask;	/* read mask for register */
     
{
  u_short tmp_val = 0;
  u_short amcc_dummy = 0;

  if (icp.cap[board]->amcc0_bug) {
      if (icp_check_amcc (board, icp.map[board]->v_cr_base + offset /
			  sizeof(u_short), ICP_AMCC_READ)) {
	if (!icp.map[board]->regs_io) {
	  ICP_DELAY(icp.cap[board]->mem_wait16);
	  amcc_dummy = icp_readw (icp.map[board]->v_cr_base + offset /
				  sizeof(u_short));
	} else {
	  ICP_DELAY(icp.cap[board]->io_wait16);
	  amcc_dummy = icp_inw (icp.map[board]->io_cr_base + offset);
	}
      }
  }
  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait16);
    }
    tmp_val = icp_readw (icp.map[board]->v_cr_base + offset /
			 sizeof(u_short));
  } else {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait16);
    }
    tmp_val = icp_inw (icp.map[board]->io_cr_base + offset);
  }

  tmp_val &= mask;
  tmp_val >>= shift;

#ifdef DEBUG_ICPCI_REG
  if (!icp.map[board]->regs_io)
    icp_printf("icpci[%d]:  cr-readw  %s, %#1X &(%#06X) >>%2d\n",
	       board, cr_reg[offset / sizeof(u_short)], tmp_val, mask, shift);
  else
    icp_printf("icpci[%d]:  cr-inw    %s, %#1X &(%#06X) >>%2d\n",
	       board, cr_reg[offset / sizeof(u_short)], tmp_val, mask, shift);
#endif

  return ((u_char)tmp_val);
}
  
/*
 * General Write Handler for Interface Control Registers
 */

/* 16 bit register access with mask and range checking */

static int
cr_write_mm (board, offset, mask, min, max, value)

  u_char board;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_short mask;		/* read mask for register */
  u_short min;		/* minimal value for range checking */
  u_short max;		/* maximal value for range checking */
  u_short value;	/* value to set to */

{
  u_short tmp_val = 0;
  u_short amcc_dummy = 0;

  if (value & ~mask) {
    icp.errno[board] = -EICP_MASK;
    return (ERROR);
  }
    
  if (value < min || value > max) {
    icp.errno[board] = -EICP_RANGE;  
    return (ERROR);
  }

  tmp_val = value & mask;

#ifdef DEBUG_ICPCI_REG
  if (!icp.map[board]->regs_io)
    icp_printf("icpci[%d]:  cr-writew %s, %#06X &(%#06X) >%#06X <%#06X\n",
	       board, cr_reg[offset / sizeof(u_short)],
	       tmp_val, mask, min, max);
  else
    icp_printf("icpci[%d]:  cr-outw   %s, %#06X &(%#06X) >%#06X <%#06X\n",
	       board, cr_reg[offset / sizeof(u_short)],
	       tmp_val, mask, min, max);
#endif

  if (icp.cap[board]->amcc0_bug) {
    if (icp_check_amcc (board, icp.map[board]->v_cr_base + offset /
			sizeof(u_short), ICP_AMCC_WRITE)) {
      if (!icp.map[board]->regs_io) {
	ICP_DELAY(icp.cap[board]->mem_wait16);
	amcc_dummy = icp_readw (icp.map[board]->v_cr_base + offset /
				sizeof(u_short));
      } else {
	ICP_DELAY(icp.cap[board]->io_wait16);
	amcc_dummy = icp_inw (icp.map[board]->io_cr_base + offset);
      }
    }
  }
  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait16);
    }
    icp_writew (tmp_val, icp.map[board]->v_cr_base + offset /
		sizeof(u_short));
  } else {
#if (0)
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait16);
    }
#endif
    icp_outw (tmp_val, icp.map[board]->io_cr_base + offset);
  }

  return (OK);
}

/* 16 bit register 1 (or more) bit access */

static int
cr_write_ms (board, offset, mask, shift, value)

  u_char board;		/* number of IC-PCI board attached to the computer */
  u_char offset;	/* offset to register base address pointer */
  u_short mask;		/* read mask for register */
  u_char  shift;	/* number of bits to shift right */
  u_char  value;	/* value to set to */  
  
{
  u_short read_val = 0;
  u_short write_val = 0;
  u_short tmp_val = 0;
  u_short amcc_dummy = 0;

  tmp_val = (u_short)value << shift;

  if (tmp_val & ~mask) {
    icp.errno[board] = -EICP_MASK;
    return (ERROR);
  }

  /* this have to be atomar */
  icp_splhigh (board);

  if (icp.cap[board]->amcc0_bug) {
    if (icp_check_amcc (board, icp.map[board]->v_cr_base + offset /
			sizeof(u_short), ICP_AMCC_READ)) {
      if (!icp.map[board]->regs_io) {
	ICP_DELAY(icp.cap[board]->mem_wait16);
	amcc_dummy = icp_readw (icp.map[board]->v_cr_base + offset /
				sizeof(u_short));
      } else {
	ICP_DELAY(icp.cap[board]->io_wait16);
	amcc_dummy = icp_inw (icp.map[board]->io_cr_base + offset);
      }
    }
  }
  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait16);
    }
    read_val = icp_readw (icp.map[board]->v_cr_base + offset /
			  sizeof(u_short));
  } else {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait16);
    }
    read_val = icp_inw (icp.map[board]->io_cr_base + offset);
  }
  
  read_val &= ~mask;
  write_val = read_val | tmp_val;

#ifdef DEBUG_ICPCI_REG
  if (!icp.map[board]->regs_io)
    icp_printf("icpci[%d]:  cr-writew %s, %#06X->%#06X %#1X &(%#06X) <<%2d\n",
	       board, cr_reg[offset / sizeof(u_short)],
	       read_val, write_val, tmp_val, mask, shift);
  else
    icp_printf("icpci[%d]:  cr-outw   %s, %#06X->%#06X %#1X &(%#06X) <<%2d\n",
	       board, cr_reg[offset / sizeof(u_short)],
	       read_val, write_val, tmp_val, mask, shift);
#endif

  if (icp.cap[board]->amcc0_bug) {
    if (icp_check_amcc (board, icp.map[board]->v_cr_base + offset /
			sizeof(u_short), ICP_AMCC_WRITE)) {
      if (!icp.map[board]->regs_io) {
	ICP_DELAY(icp.cap[board]->mem_wait16);
	amcc_dummy = icp_readw (icp.map[board]->v_cr_base + offset /
				sizeof(u_short));
      } else {
	ICP_DELAY(icp.cap[board]->io_wait16);
	amcc_dummy = icp_inw (icp.map[board]->io_cr_base + offset);
      }
    }
  }
  if (!icp.map[board]->regs_io) {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->mem_wait16);
    }
    icp_writew (write_val, icp.map[board]->v_cr_base + offset /
		sizeof(u_short));
  } else {
    if (icp.cap[board]->amcc1_bug) {
      ICP_DELAY(icp.cap[board]->io_wait16);
    }
    icp_outw (write_val, icp.map[board]->io_cr_base + offset);
  }
  
  icp_spllow (board);	/* end atomar */
  
  return (OK);
}
  
/*
 * Special Read Handler for IC-PCI Control Registers
 */

u_char
cr_get_bmstren (u_char board) {

  return (cr_read_ms (board,
		      CR_CON_OFFSET,
		      CR_BMSTREN_MASK,
		      CR_BMSTREN_SHIFT));
}

u_char
cr_get_bmilace (u_char board) {

  return (cr_read_ms (board,
		      CR_CON_OFFSET,
		      CR_BMILACE_MASK,
		      CR_BMILACE_SHIFT));
}

u_char
cr_get_pixsz (u_char board) {

  u_char tmp_val;

  tmp_val = cr_read_ms (board,
		      CR_CON_OFFSET,
		      CR_PIXSZ_MASK,
		      CR_PIXSZ_SHIFT);

#if(0)
  if (icp.pci[board].revision < ICP_REV_B &&tmp_val == 3)
    icp.errno[board] = -EICP_RANGE;
#endif

  return (tmp_val);
}

u_char
cr_get_memcfg (u_char board) {

  return (cr_read_ms (board,
		      CR_CON_OFFSET,
		      CR_MEMCFG_MASK,
		      CR_MEMCFG_SHIFT));
}

u_char
cr_get_vbstat (u_char board) {

  return (cr_read_ms (board,
		      CR_CON_OFFSET,
		      CR_VBSTAT_MASK,
		      CR_VBSTAT_SHIFT));
}

u_char
cr_get_fldstat (u_char board) {

  return (cr_read_ms (board,
		      CR_CON_OFFSET,
		      CR_FLDSTAT_MASK,
		      CR_FLDSTAT_SHIFT));
}

u_char
cr_get_smstat (u_char board) {

  return (cr_read_ms (board,
		      CR_CON_OFFSET,
		      CR_SMSTAT_MASK,
		      CR_SMSTAT_SHIFT));
}

u_char
cr_get_amp (u_char board) {

  return (cr_read_ms (board,
		      CR_CON_OFFSET,
		      CR_AMP_MASK,
		      CR_AMP_SHIFT));
}

u_char
cr_get_tmode (u_char board) {

  return (cr_read_ms (board,
		      CR_ACQCON_OFFSET,
		      CR_TMODE_MASK,
		      CR_TMODE_SHIFT));
}

u_char
cr_get_fcnt (u_char board) {

  return (cr_read_ms (board,
		      CR_ACQCON_OFFSET,
		      CR_FCNT_MASK,
		      CR_FCNT_SHIFT) + 1);
}

u_char
cr_get_fldsel (u_char board) {

  u_char tmp_val;

  tmp_val = cr_read_ms (board,
			CR_ACQCON_OFFSET,
			CR_FLDSEL_MASK,
			CR_FLDSEL_SHIFT);

#if(0)
  if (tmp_val == 1) icp.errno[board] = -EICP_RANGE;
#endif

  return (tmp_val);
}

u_char
cr_get_acqmd (u_char board) {

  u_char tmp_val;

  tmp_val = cr_read_ms (board,
			CR_ACQCON_OFFSET,
			CR_ACQMD_MASK,
			CR_ACQMD_SHIFT);

#if(0)
  if (tmp_val == 1) icp.errno[board] = -EICP_RANGE;
#endif

  return (tmp_val);
}

u_char
cr_get_newaq (u_char board) {

  return (cr_read_ms (board,
		      CR_ACQCON_OFFSET,
		      CR_NEWAQ_MASK,
		      CR_NEWAQ_SHIFT));
}

u_char
cr_get_gstat (u_char board) {

  return (cr_read_ms (board,
		      CR_ACQCON_OFFSET,
		      CR_GSTAT_MASK,
		      CR_GSTAT_SHIFT));
}

u_char
cr_get_fstrt (u_char board) {

  return (cr_read_ms (board,
		      CR_ACQCON_OFFSET,
		      CR_FSTRT_MASK,
		      CR_FSTRT_SHIFT));
}

u_short
cr_get_pitch (u_char board) {

  u_short tmp_val;

  tmp_val = cr_read_mm (board,
			CR_PITCH_OFFSET,
			CR_PITCH_MASK,
			CR_PITCH_MIN,
			CR_PITCH_MAX);

  return (tmp_val << CR_PITCH_SHIFT);
}

u_short
cr_get_aoix (u_char board) {

  u_short tmp_val;

  tmp_val =  cr_read_mm (board,
			 CR_AOIX_OFFSET,
			 CR_AOIX_MASK,
			 CR_AOIX_MIN,
			 CR_AOIX_MAX);

  return (tmp_val << CR_AOIX_SHIFT);
}

u_long
cr_get_bmsf (u_char board) {

  u_short tmp_low, tmp_high;

  tmp_low = cr_read_mm (board,
			CR_BMSFL_OFFSET,
			CR_BMSFL_MASK,
			CR_BMSFL_MIN,
			CR_BMSFL_MAX);

  if (icp.map[board]->big_mem) {
    tmp_high = cr_read_mm (board,
			   CR_BMSFH_OFFSET,
			   (CR_BMSFH_MASK << 1) + 1,
			   CR_BMSFH_MIN,
			   (CR_BMSFH_MAX << 1) + 1);
  } else {
    tmp_high = cr_read_mm (board,
			   CR_BMSFH_OFFSET,
			   CR_BMSFH_MASK,
			   CR_BMSFH_MIN,
			   CR_BMSFH_MAX);
  }

  tmp_low &= CR_BMSFL_MASK;

  if (icp.map[board]->big_mem) {
    tmp_high &= (CR_BMSFH_MASK << 1) + 1;    
  } else {
    tmp_high &= CR_BMSFH_MASK;
  }

  return ((tmp_low << CR_BMSFL_SHIFT) | ((u_long)tmp_high << CR_BMSFH_SHIFT));
}

u_long
cr_get_bmss (u_char board) {

  u_short tmp_low, tmp_high;

  tmp_low = cr_read_mm (board,
			CR_BMSSL_OFFSET,
			CR_BMSSL_MASK,
			CR_BMSSL_MIN,
			CR_BMSSL_MAX);

  if (icp.map[board]->big_mem) {
    tmp_high = cr_read_mm (board,
			   CR_BMSSH_OFFSET,
			   (CR_BMSSH_MASK << 1) + 1,
			   CR_BMSSH_MIN,
			   (CR_BMSSH_MAX << 1) + 1);
  } else {
    tmp_high = cr_read_mm (board,
			   CR_BMSSH_OFFSET,
			   CR_BMSSH_MASK,
			   CR_BMSSH_MIN,
			   CR_BMSSH_MAX);
  }

  tmp_low &= CR_BMSSL_MASK;
  
  if (icp.map[board]->big_mem) {
    tmp_high &= (CR_BMSSH_MASK << 1) + 1;    
  } else {
    tmp_high &= CR_BMSSH_MASK;
  }
  
  return ((tmp_low << CR_BMSSL_SHIFT) | ((u_long)tmp_high << CR_BMSSH_SHIFT));
}

u_long
cr_get_aqstart (u_char board) {

  u_long tmp_val;
  
  if (icp.map[board]->big_mem) {
    tmp_val = cr_read_mm (board,
			  CR_AQSTART_OFFSET,
			  (CR_AQSTART_MASK << 1) + 1,
			  CR_AQSTART_MIN,
			  (CR_AQSTART_MAX << 1) + 1);
  } else {
    tmp_val = cr_read_mm (board,
			  CR_AQSTART_OFFSET,
			  CR_AQSTART_MASK,
			  CR_AQSTART_MIN,
			  CR_AQSTART_MAX);
  }

  return (tmp_val << CR_AQSTART_SHIFT);
}

u_long
cr_get_aqadr (u_char board) {

  u_long tmp_val;

  if (icp.map[board]->big_mem) {
    tmp_val = cr_read_mm (board,
			  CR_AQADR_OFFSET,
			  (CR_AQADR_MASK << 1) + 1,
			  CR_AQADR_MIN,
			  (CR_AQADR_MAX << 1) + 1);
  } else {
    tmp_val = cr_read_mm (board,
			  CR_AQADR_OFFSET,
			  CR_AQADR_MASK,
			  CR_AQADR_MIN,
			  CR_AQADR_MAX);
  }

  return (tmp_val << CR_AQADR_SHIFT);
}

u_long
cr_get_intadr (u_char board) {

  u_long tmp_val;

  if (icp.map[board]->big_mem) {
    tmp_val = cr_read_mm (board,
			  CR_INTADR_OFFSET,
			  (CR_INTADR_MASK << 1) + 1,
			  CR_INTADR_MIN,
			  (CR_INTADR_MAX << 1) + 1);
  } else {
    tmp_val = cr_read_mm (board,
			  CR_INTADR_OFFSET,
			  CR_INTADR_MASK,
			  CR_INTADR_MIN,
			  CR_INTADR_MAX);
  }
  return (tmp_val << CR_INTADR_SHIFT);
}

u_char
cr_get_vbien (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_VBIEN_MASK,
		      CR_VBIEN_SHIFT));
}

u_char
cr_get_ovbien (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_OVBIEN_MASK,
		      CR_OVBIEN_SHIFT));
}

u_char
cr_get_evbien (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_EVBIEN_MASK,
		      CR_EVBIEN_SHIFT));
}

u_char
cr_get_amien (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_AMIEN_MASK,
		      CR_AMIEN_SHIFT));
}

u_char
cr_get_trigien (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_TRIGIEN_MASK,
		      CR_TRIGIEN_SHIFT));
}

u_char
cr_get_adrien (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_ADRIEN_MASK,
		      CR_ADRIEN_SHIFT));
}

u_char
cr_get_aqien (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_AQIEN_MASK,
		      CR_AQIEN_SHIFT));
}

/* old, use xcr_test_irq */
u_char
cr_get_vbint (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_VBINT_MASK,
		      CR_VBINT_SHIFT));
}

/* old, use xcr_test_irq */
u_char
cr_get_ovbint (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_OVBINT_MASK,
		      CR_OVBINT_SHIFT));
}

/* old, use xcr_test_irq */
u_char
cr_get_evbint (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_EVBINT_MASK,
		      CR_EVBINT_SHIFT));
}

/* old, use xcr_test_irq */
u_char
cr_get_amint (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_AMINT_MASK,
		      CR_AMINT_SHIFT));
}

/* old, use xcr_test_irq */
u_char
cr_get_trigint (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_TRIGINT_MASK,
		      CR_TRIGINT_SHIFT));
}

/* old, use xcr_test_irq */
u_char
cr_get_adrint (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_ADRINT_MASK,
		      CR_ADRINT_SHIFT));
}

/* old, use xcr_test_irq */
u_char
cr_get_aqint (u_char board) {

  return (cr_read_ms (board,
		      CR_INTCON_OFFSET,
		      CR_AQINT_MASK,
		      CR_AQINT_SHIFT));
}

/* REV-B Only! */
u_char
cr_get_hdec (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_WZOOM_OFFSET,
		      CR_HDEC_MASK,
		      CR_HDEC_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_vdec (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_WZOOM_OFFSET,
		      CR_VDEC_MASK,
		      CR_VDEC_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_clip (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
 
  return (cr_read_ms (board,
		      CR_WZOOM_OFFSET,
		      CR_CLIP_MASK,
		      CR_CLIP_SHIFT));  
}

/* PCVision Only! */
u_char
cr_get_bmzoom (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B ||
      icp.cap[board]->module != ICP_AMPV)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_WZOOM_OFFSET,
		      CR_BMZOOM_MASK,
		      CR_BMZOOM_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_outstrb (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_OUTSTRB_MASK,
		      CR_OUTSTRB_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_inregenb (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_INREG_ENB_MASK,
		      CR_INREG_ENB_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_instrbpol (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_INSTRB_POL_MASK,
		      CR_INSTRB_POL_SHIFT));  

}

/* REV-B Only! */
u_char
cr_get_inregclr (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_INREG_CLR_MASK,
		      CR_INREG_CLR_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_inportien (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_INPORT_IEN_MASK,
		      CR_INPORT_IEN_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_inportpol (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_INPORT_POL_MASK,
		      CR_INPORT_POL_SHIFT));  
}

/* REV-B Only! old, use xcr_test_irq */
u_char
cr_get_inportint (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_INPORT_INT_MASK,
		      CR_INPORT_INT_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_inportrint (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_INPORT_RINT_MASK,
		      CR_INPORT_RINT_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_inportrstb (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_INPORT_RSTB_MASK,
		      CR_INPORT_RSTB_SHIFT));  
}

/* PCVision Only! */
u_char
cr_get_pad16 (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B ||
      icp.cap[board]->module != ICP_AMPV)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_PAD16_MASK,
		      CR_PAD16_SHIFT));  
}

/* REV-B Only! */
u_char
cr_get_outport (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }

  return (cr_read_mm (board,
		      CR_OUTPORT_OFFSET,
		      CR_OUTPORT_MASK,
		      CR_OUTPORT_MIN,
		      CR_OUTPORT_MAX));
}

/* REV-B Only! */
u_char
cr_get_inport (u_char board)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_read_mm (board,
		      CR_INPORT_OFFSET,
		      CR_INPORT_MASK,
		      CR_INPORT_MIN,
		      CR_INPORT_MAX));
}

/*
 * Special Write Handler for IC-PCI Control Registers
 */

int
cr_set_bmstren (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_CON_OFFSET,
		       CR_BMSTREN_MASK,
		       CR_BMSTREN_SHIFT,
		       value));
}
  
int
cr_set_bmilace (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_CON_OFFSET,
		       CR_BMILACE_MASK,
		       CR_BMILACE_SHIFT,
		       value));
}
  
int
cr_set_pixsz (u_char board, u_char value) {

  if (icp.pci[board]->revision < ICP_REV_B && value == 3) {
    icp.errno[board] = -EICP_RANGE;
    return (ERROR);
  }

  return (cr_write_ms (board,
		       CR_CON_OFFSET,
		       CR_PIXSZ_MASK,
		       CR_PIXSZ_SHIFT,
		       value));
}
  
int
cr_set_memcfg (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_CON_OFFSET,
		       CR_MEMCFG_MASK,
		       CR_MEMCFG_SHIFT,
		       value));
}

int
cr_set_tmode (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_ACQCON_OFFSET,
		       CR_TMODE_MASK,
		       CR_TMODE_SHIFT,
		       value));
}
  
int
cr_set_fcnt (u_char board, u_char value) {

  if (--value > (CR_FCNT_MASK >> CR_FCNT_SHIFT)) {
    icp.errno[board] = -EICP_RANGE;
    return (ERROR);
  }

  return (cr_write_ms (board,
		       CR_ACQCON_OFFSET,
		       CR_FCNT_MASK,
		       CR_FCNT_SHIFT,
		       value));
}
  
int
cr_set_fldsel (u_char board, u_char value) {

  if (value == 1) {
    icp.errno[board] = -EICP_RANGE;
    return (ERROR);
  }

    return (cr_write_ms (board,
			 CR_ACQCON_OFFSET,
			 CR_FLDSEL_MASK,
			 CR_FLDSEL_SHIFT,
			 value));
	    }
  
int
cr_set_acqmd (u_char board, u_char value) {

  if (value == 1) {
    icp.errno[board] = -EICP_RANGE;
    return (ERROR);
  }

    return (cr_write_ms (board,
			 CR_ACQCON_OFFSET,
			 CR_ACQMD_MASK,
			 CR_ACQMD_SHIFT,
			 value));
}
  
int
cr_set_pitch (u_char board, u_short value) {

  u_short tmp_val;

  if (value & ((1 << CR_PITCH_SHIFT) - 1))
    {
      icp.errno[board] = -EICP_MASK;
      return (ERROR);
    }
  
  tmp_val = value >> CR_PITCH_SHIFT;

  return (cr_write_mm (board,
		       CR_PITCH_OFFSET,
		       CR_PITCH_MASK,
		       CR_PITCH_MIN,
		       CR_PITCH_MAX,
		       tmp_val));
}

int
cr_set_aoix (u_char board, u_short value) {

  u_short tmp_val;

  if (value & ((1 << CR_AOIX_SHIFT) - 1))
    {
      icp.errno[board] = -EICP_MASK;
      return (ERROR);
    }

  tmp_val = value >> CR_AOIX_SHIFT;

  return (cr_write_mm (board,
		       CR_AOIX_OFFSET,
		       CR_AOIX_MASK,
		       CR_AOIX_MIN,
		       CR_AOIX_MAX,
		       tmp_val));
}

int
cr_set_bmsf (u_char board, u_long value) {

  u_short tmp_low, tmp_high;

  if (value & ((1 << CR_BMSFL_SHIFT) - 1))
    {
      icp.errno[board] = -EICP_MASK;
      return (ERROR);
    }

  tmp_low = (value >> CR_BMSFL_SHIFT) & CR_BMSFL_MASK;

  if (icp.map[board]->big_mem) {
    tmp_high = (value >> CR_BMSFH_SHIFT) & ((CR_BMSFH_MASK << 1) + 1);
  } else {
    tmp_high = (value >> CR_BMSFH_SHIFT) & CR_BMSFH_MASK;
  }

  if (cr_write_mm (board,
		   CR_BMSFL_OFFSET,
		   CR_BMSFL_MASK,
		   CR_BMSFL_MIN,
		   CR_BMSFL_MAX,
		   tmp_low) < 0) return (ERROR);

  if (icp.map[board]->big_mem) {
    return (cr_write_mm (board,
			 CR_BMSFH_OFFSET,
			 (CR_BMSFH_MASK << 1) + 1,
			 CR_BMSFH_MIN,
			 (CR_BMSFH_MAX << 1) + 1,
			 tmp_high));
  } else {
    return (cr_write_mm (board,
			 CR_BMSFH_OFFSET,
			 CR_BMSFH_MASK,
			 CR_BMSFH_MIN,
			 CR_BMSFH_MAX,
			 tmp_high));
  }
}

int
cr_set_bmss (u_char board, u_long value) {

  u_short tmp_low, tmp_high;

  if (value & ((1 << CR_BMSSL_SHIFT) - 1))
    {
      icp.errno[board] = -EICP_MASK;
      return (ERROR);
    }

  tmp_low = (value >> CR_BMSSL_SHIFT) & CR_BMSSL_MASK;

  if (icp.map[board]->big_mem) {
    tmp_high = (value >> CR_BMSSH_SHIFT) & ((CR_BMSSH_MASK << 1) + 1);
  } else {
    tmp_high = (value >> CR_BMSSH_SHIFT) & CR_BMSSH_MASK;
  }

  if (cr_write_mm (board,
		   CR_BMSSL_OFFSET,
		   CR_BMSSL_MASK,
		   CR_BMSSL_MIN,
		   CR_BMSSL_MAX,
		   tmp_low) < 0) return (ERROR);

  if (icp.map[board]->big_mem) {
    return (cr_write_mm (board,
			 CR_BMSSH_OFFSET,
			 (CR_BMSSH_MASK << 1) + 1,
			 CR_BMSSH_MIN,
			 (CR_BMSSH_MAX << 1) + 1,
			 tmp_high));
  } else {
    return (cr_write_mm (board,
			 CR_BMSSH_OFFSET,
			 CR_BMSSH_MASK,
			 CR_BMSSH_MIN,
			 CR_BMSSH_MAX,
			 tmp_high));
  }
  return (OK);
}

int
cr_set_aqstart (u_char board, u_long value) {

  u_long tmp_val;

  if (value & ((1 << CR_AQSTART_SHIFT) - 1))
    {
      icp.errno[board] = -EICP_MASK;
      return (ERROR);
    }

  tmp_val = value >> CR_AQSTART_SHIFT;

  if (icp.map[board]->big_mem) {
    return (cr_write_mm (board,
			 CR_AQSTART_OFFSET,
			 (CR_AQSTART_MASK << 1) + 1,
			 CR_AQSTART_MIN,
			 (CR_AQSTART_MAX << 1) + 1,
			 (u_short)tmp_val));
  } else {
    return (cr_write_mm (board,
			 CR_AQSTART_OFFSET,
			 CR_AQSTART_MASK,
			 CR_AQSTART_MIN,
			 CR_AQSTART_MAX,
			 (u_short)tmp_val));
  }
  return (OK);
}

int
cr_set_intadr (u_char board, u_long value) {

  u_long tmp_val;

  if (value & ((1 << CR_INTADR_SHIFT) - 1))
    {
      icp.errno[board] = -EICP_MASK;
      return (ERROR);
    }

  tmp_val = value >> CR_INTADR_SHIFT;

  if (icp.map[board]->big_mem) {
    return (cr_write_mm (board,
			 CR_INTADR_OFFSET,
			 (CR_INTADR_MASK << 1) + 1,
			 CR_INTADR_MIN,
			 (CR_INTADR_MAX << 1) + 1,
			 (u_short)tmp_val));
  } else {
    return (cr_write_mm (board,
			 CR_INTADR_OFFSET,
			 CR_INTADR_MASK,
			 CR_INTADR_MIN,
			 CR_INTADR_MAX,
			 (u_short)tmp_val));
  }
}

/* old, use icp_[en|dis]able_irq */
int
cr_set_vbien (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_INTCON_OFFSET,
		       CR_VBIEN_MASK,
		       CR_VBIEN_SHIFT,
		       value));
}

/* old, use icp_[en|dis]able_irq */
int
cr_set_ovbien (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_INTCON_OFFSET,
		       CR_OVBIEN_MASK,
		       CR_OVBIEN_SHIFT,
		       value));
}

/* old, use icp_[en|dis]able_irq */
int
cr_set_evbien (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_INTCON_OFFSET,
		       CR_EVBIEN_MASK,
		       CR_EVBIEN_SHIFT,
		       value));
}

/* old, use icp_[en|dis]able_irq */
int
cr_set_amien (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_INTCON_OFFSET,
		       CR_AMIEN_MASK,
		       CR_AMIEN_SHIFT,
		       value));
}

/* old, use icp_[en|dis]able_irq */
int
cr_set_trigien (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_INTCON_OFFSET,
		       CR_TRIGIEN_MASK,
		       CR_TRIGIEN_SHIFT,
		       value));
}

/* old, use icp_[en|dis]able_irq */
int
cr_set_adrien (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_INTCON_OFFSET,
		       CR_ADRIEN_MASK,
		       CR_ADRIEN_SHIFT,
		       value));
}

/* old, use icp_[en|dis]able_irq */
int
cr_set_aqien (u_char board, u_char value) {

  return (cr_write_ms (board,
		       CR_INTCON_OFFSET,
		       CR_AQIEN_MASK,
		       CR_AQIEN_SHIFT,
		       value));
}

/* REV-B Only! */   
int
cr_set_hdec (u_char board, u_char value)
{

  int bpp;

  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  switch (cr_get_pixsz (board))
    {
    case CR_PIXSZ_8:
      bpp = 1;
      break;
    case CR_PIXSZ_16:
      bpp = 2;
      break;
    case CR_PIXSZ_24_24:
      bpp = 4;
      break;
    default:
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  switch (value)
    {
    case CR_DEC_BY_1:
      if (icp.mem[board]->width % (8 * bpp))
	{
	  icp.errno[board] = -EICP_MASK;
	  return (ERROR);
	}
      break;
    case CR_DEC_BY_2:
      if (icp.mem[board]->width % (16 * bpp))
	{
	  icp.errno[board] = -EICP_MASK;
	  return (ERROR);
	}
      break;
    case CR_DEC_BY_4:
      if (icp.mem[board]->width % (32 * bpp))
	{
	  icp.errno[board] = -EICP_MASK;
	  return (ERROR);
	}
      break;
    case CR_DEC_BY_8:
      if (icp.mem[board]->width % (64 * bpp))
	{
	  icp.errno[board] = -EICP_MASK;
	  return (ERROR);
	}
      if (bpp != 1)	/* this is only supported for 8bit images */
	{
	  icp.errno[board] = -EICP_NOSUPPORT;
	  return (ERROR);
	}
      break;
    default:
      icp.errno[board] = -EICP_RANGE;
      return (ERROR);
    }
  
  return (cr_write_ms (board,
		       CR_WZOOM_OFFSET,
		       CR_HDEC_MASK,
		       CR_HDEC_SHIFT,
		       value));
}	

/* REV-B Only! */   
int
cr_set_vdec (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  switch (value)
    {
    case CR_DEC_BY_1:
    case CR_DEC_BY_2:
    case CR_DEC_BY_4:
    case CR_DEC_BY_8:
      break;
    default:
      icp.errno[board] = -EICP_RANGE;
      return (ERROR);
    }

  return (cr_write_ms (board,
		       CR_WZOOM_OFFSET,
		       CR_VDEC_MASK,
		       CR_VDEC_SHIFT,
		       value));
}	

/* REV-B Only! */   
int
cr_set_clip (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_write_ms (board,
		       CR_WZOOM_OFFSET,
		       CR_CLIP_MASK,
		       CR_CLIP_SHIFT,
		       value));
}	

/* PCVision Only! */
int
cr_set_bmzoom (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B ||
      icp.cap[board]->module != ICP_AMPV)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_write_ms (board,
		       CR_WZOOM_OFFSET,
		       CR_BMZOOM_MASK,
		       CR_BMZOOM_SHIFT,
		       value));
}	

/* REV-B Only! */   
int 
cr_set_outstrb (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }

  return (cr_write_ms (board,
		       CR_PORTCON_OFFSET,
		       CR_OUTSTRB_MASK,
		       CR_OUTSTRB_SHIFT,
		       value));
 
}	

/* REV-B Only! */   
int 
cr_set_inregenb (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_write_ms (board,
		       CR_PORTCON_OFFSET,
		       CR_INREG_ENB_MASK,
		       CR_INREG_ENB_SHIFT,
		       value));
}
	
/* REV-B Only! */   
int
cr_set_instrbpol (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_write_ms (board,
		       CR_PORTCON_OFFSET,
		       CR_INSTRB_POL_MASK,
		       CR_INSTRB_POL_SHIFT,
		       value));

}	

/* REV-B Only! */   
int
cr_set_inregclr (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_write_ms (board,
		       CR_PORTCON_OFFSET,
		       CR_INREG_CLR_MASK,
		       CR_INREG_CLR_SHIFT,
		       value));

}	

/* REV-B Only! */   
int
cr_set_inportien (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_write_ms (board,
		       CR_PORTCON_OFFSET,
		       CR_INPORT_IEN_MASK,
		       CR_INPORT_IEN_SHIFT,
		       value));

}
	
/* REV-B Only! */   
int
cr_set_inportpol (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_write_ms (board,
		       CR_PORTCON_OFFSET,
		       CR_INPORT_POL_MASK,
		       CR_INPORT_POL_SHIFT,
		       value));
}	

/* PCVision Only! */
int
cr_set_pad16 (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B ||
      icp.cap[board]->module != ICP_AMPV)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }
  
  return (cr_write_ms (board,
		       CR_PORTCON_OFFSET,
		       CR_PAD16_MASK,
		       CR_PAD16_SHIFT,
		       value));
}	

/* REV-B Only! */   
int
cr_set_outport (u_char board, u_char value)
{
  if (icp.pci[board]->revision < ICP_REV_B)
    {
      icp.errno[board] = -EICP_NOSUPPORT;
      return (ERROR);
    }

  return (cr_write_mm (board,
		       CR_OUTPORT_OFFSET,
		       CR_OUTPORT_MASK,
		       CR_OUTPORT_MIN,
		       CR_OUTPORT_MAX,
		       (u_short)value));
}	

/*
 * Special Initialization Handler for IC-PCI Bitfield Control Registers
 */

int
cr_init_con (u_char board) {

  return (cr_write_mm (board,
		       CR_CON_OFFSET,
		       CR_NO_MASK,
		       CR_NO_MIN,
		       CR_NO_MAX,
		       CR_CON_INIT));
}
		
int
cr_init_acqcon (u_char board) {

  return (cr_write_mm (board,
		       CR_ACQCON_OFFSET,
		       CR_NO_MASK,
		       CR_NO_MIN,
		       CR_NO_MAX,
		       CR_ACQCON_INIT));
}

int
cr_init_intcon (u_char board) {

  return (cr_write_mm (board,
		       CR_INTCON_OFFSET,
		       CR_NO_MASK,
		       CR_NO_MIN,
		       CR_NO_MAX,
		       CR_INTCON_INIT));
}

int
cr_init_wzoom (u_char board)
{
  
  return (cr_write_mm (board,
		       CR_WZOOM_OFFSET,
		       CR_NO_MASK,
		       CR_NO_MIN,
		       CR_NO_MAX,
		       CR_WZOOM_INIT));
}

int
cr_init_mdata (u_char board) {
  
  return (cr_write_mm (board,
		       CR_MDATA_OFFSET,
		       CR_NO_MASK,
		       CR_NO_MIN,
		       CR_NO_MAX,
		       CR_MDATA_INIT));
}

int
cr_init_portcon (u_char board)
{
  
  return (cr_write_mm (board,
		       CR_PORTCON_OFFSET,
		       CR_NO_MASK,
		       CR_NO_MIN,
		       CR_NO_MAX,
		       CR_PORTCON_INIT));
}

int
cr_init_outport (u_char board) {

  return (cr_write_mm (board,
		       CR_OUTPORT_OFFSET,
		       CR_NO_MASK,
		       CR_NO_MIN,
		       CR_NO_MAX,
		       CR_OUTPORT_INIT));
}

int
cr_init_inport (u_char board)
{

  return (cr_write_mm (board,
		       CR_INPORT_OFFSET,
		       CR_NO_MASK,
		       CR_NO_MIN,
		       CR_NO_MAX,
		       CR_INPORT_INIT));
}

/*
 * Special Interrupt clear function (called by icp_intr/icp_en|disable_irq)
 */

void
xcr_clear_irqs (u_char board)
{

  u_short cr_additional = 0;
  u_short cr_standard = 0;

  /* clear additional PCVision interrupts */
  if (icp.cap[board]->module == ICP_AMPV)
    icp_writew (icp_readw (&pv_reg[board]->con2.all) | 0x0F00,
		&pv_reg[board]->con2.all);
		
  /* mark and clear additional IC-PCI REV-B and PCVision interrupts */
  if (icp.pci[board]->revision > ICP_REV_A)
    {
      if (cr_read_ms (board,
		      CR_PORTCON_OFFSET,
		      CR_INPORT_IEN_MASK,
		      CR_INPORT_IEN_SHIFT))
	cr_additional |= CR_INPINT;
      cr_write_ms (board,
		   CR_PORTCON_OFFSET,
		   CR_INPORT_IEN_MASK,
		   CR_INPORT_IEN_SHIFT,
		   CR_DISABLE);
      
    }

  /* mark and clear standard IC-PCI interrupts */
  cr_standard = cr_read_mm (board,
			    CR_INTCON_OFFSET,
			    CR_NO_MASK,
			    CR_NO_MIN,
			    CR_NO_MAX);
  cr_write_mm (board,
	       CR_INTCON_OFFSET,
	       CR_NO_MASK,
	       CR_NO_MIN,
	       CR_NO_MAX,
	       0x0000);

  /* clear main AMCC (PCI-Matchmaker) interrupts */
  icr_write_mm (board,
		ICR_INTCTL_OFFSET,
		ICR_NO_MASK,
		ICR_NO_MIN,
		ICR_NO_MAX,
		icr_read_mm (board,
			     ICR_INTCTL_OFFSET,
			     ICR_NO_MASK,
			     ICR_NO_MIN,
			     ICR_NO_MAX) | ICR_INTCTL_INIT);

  /* reenable standard IC-PCI interrupts */
  cr_write_mm (board,
	       CR_INTCON_OFFSET,
	       CR_NO_MASK,
	       CR_NO_MIN,
	       CR_NO_MAX,
	       cr_standard);

  /* reenable additional IC-PCI REV-B and PCVision interrupts */
  if (icp.pci[board]->revision > ICP_REV_A)
    if (cr_additional & CR_INPINT)
      cr_write_ms (board,
		   CR_PORTCON_OFFSET,
		   CR_INPORT_IEN_MASK,
		   CR_INPORT_IEN_SHIFT,
		   CR_ENABLE);
}

/*
 * Special Interrupt test function (called by icp_intr)
 */

void
xcr_test_irqs (u_char board, u_long *irqs)
{

  u_short pv_additional = 0;
  u_short cr_additional = 0;
  u_short cr_standard = 0;
  u_long icr_standard = 0UL;
  
  /* test additional PCVision interrupts */
  if (icp.cap[board]->module == ICP_AMPV)
    {
      pv_additional = icp_readw (&pv_reg[board]->con2.all);
      pv_additional &= PV_IRQ_TEST_MASK;
      pv_additional >>= PV_IRQ_TEST_SHIFT;
    }

  /* test additional IC-PCI REV-B and PCVision interrupts */
  if (icp.pci[board]->revision > ICP_REV_A)
    if (cr_read_ms (board,
		    CR_PORTCON_OFFSET,
		    CR_INPORT_INT_MASK,
		    CR_INPORT_INT_SHIFT))
      cr_additional |= CR_INPINT;

  /* test standard IC-PCI interrupts */  
  cr_standard = cr_read_mm (board,
			    CR_INTCON_OFFSET,
			    CR_NO_MASK,
			    CR_NO_MIN,
			    CR_NO_MAX);
  cr_standard &= CR_IRQ_TEST_MASK;

  /* test main AMCC (PCI-Matchmaker) interrupts */  
  icr_standard = icr_read_mm (board,
			      ICR_INTCTL_OFFSET,
			      ICR_NO_MASK,
			      ICR_NO_MIN,
			      ICR_NO_MAX);
  icr_standard &= ICR_IRQ_TEST_MASK;

  *irqs = icr_standard | cr_standard | cr_additional | pv_additional;
}

